/*
   Watch6 ARM7 Code
   Maarten Pennings - 2008 jan 5
   From Watch5, but merged arm7 template from nds-examples-20071023
   Added power management (backlight control and system power).
*/

#include <nds.h>


enum Commands { cNone, cPowerOff, cLightBotOn, cLightBotOff, cLightTopOn, cLightTopOff };
#define COMMAND (IPC->mailData)


/* Protocol (Communication & Synchronisation) with Arm9 on sound
 IPC->aux expresses which sound channels are busy, i.e. bit n equals SCHANNEL_ENABLE of SCHANNEL_CR(n)
 Note that IPC->aux is updated by VcountHandler, so it lags behind.
 Therefore, when manipulating SCHANNEL_CR, also IPC->aux is updated.
*/
#define ARM7_ALTBITS     (IPC->battery         )
#define ARM7_ALTBIT(n)   (IPC->battery & BIT(n))
#define ARM7_BUSYBITS    (IPC->aux             )
#define ARM7_BUSYBIT(n)  (IPC->aux     & BIT(n))

//---------------------------------------------------------------------------------
void startSound(int sampleRate, const void* data, u32 bytes, u8 channel, u8 vol,  u8 pan, u8 format) {
//---------------------------------------------------------------------------------
  SCHANNEL_TIMER(channel)  = SOUND_FREQ(sampleRate);
  SCHANNEL_SOURCE(channel) = (u32)data;
  SCHANNEL_LENGTH(channel) = bytes >> 2 ;
  SCHANNEL_CR(channel)     = SCHANNEL_ENABLE | SOUND_ONE_SHOT | SOUND_VOL(vol) | SOUND_PAN(pan) | (format==1?SOUND_8BIT:SOUND_16BIT);
  // Arm9 signalling
  ARM7_BUSYBITS |= BIT(channel);  // Arm7 is busy on 'channel'
  ARM7_ALTBITS  ^= BIT(channel);  // Signal (alternating bit) that the Arm7 processed the 'Arm9 start sound command'
}


touchPosition first,tempPos;

//---------------------------------------------------------------------------------
void VcountHandler() {
//---------------------------------------------------------------------------------
  static int lastbut = -1;
  uint16 but=0, x=0, y=0, xpx=0, ypx=0, z1=0, z2=0;
  uint16 sndBusy;
  int i;

  but = REG_KEYXY;

  // Get the busy flag of all sound channels
  sndBusy=0;
  for( i=0; i<16; i++ )
    if( SCHANNEL_CR(i)&SCHANNEL_ENABLE ) sndBusy|=BIT(i);
  ARM7_BUSYBITS=sndBusy;

  if (!( (but ^ lastbut) & (1<<6))) {

    tempPos = touchReadXY();

    if ( tempPos.x == 0 || tempPos.y == 0 ) {
      but |= (1 <<6);
      lastbut = but;
    } else {
      x = tempPos.x;
      y = tempPos.y;
      xpx = tempPos.px;
      ypx = tempPos.py;
      z1 = tempPos.z1;
      z2 = tempPos.z2;
    }

  } else {
    lastbut = but;
    but |= (1 <<6);
  }

  IPC->touchX     = x;
  IPC->touchY     = y;
  IPC->touchXpx   = xpx;
  IPC->touchYpx   = ypx;
  IPC->touchZ1    = z1;
  IPC->touchZ2    = z2;
  IPC->buttons    = but;

}

//---------------------------------------------------------------------------------
void VblankHandler(void) {
//---------------------------------------------------------------------------------
  u32 i;
  TransferSound *snd = IPC->soundData;
  IPC->soundData = 0; // Flag (to our own next Vblank invocation) that we did startSound

  if( 0!=snd ) {
    if( snd->count>16 ) snd->count=16; // buffer overflow protection
    for( i=0; i<snd->count; i++ ) {
      s32 chan = i; // new style: TransferSound at index i is for sound channel i
      if( snd->data[i].len>0 ) {
        startSound(snd->data[i].rate, snd->data[i].data, snd->data[i].len, chan, snd->data[i].vol, snd->data[i].pan, snd->data[i].format);
      }
    }
  }
}

//---------------------------------------------------------------------------------
int main(int argc, char ** argv) {
//---------------------------------------------------------------------------------
  // read User Settings from firmware
  readUserSettings();

  // Arm9 signalling
  ARM7_BUSYBITS = 0; // Arm7 is busy with none of the channels
  ARM7_ALTBITS  = 0; // Arbitrary 'alternating bit' settings for all channels (all 0) -- must be the same on Arm9

  IPC->soundData = 0;

  // Reset the clock if needed
  rtcReset();

  //enable sound
  powerON(POWER_SOUND);
  writePowerManagement(PM_CONTROL_REG, ( readPowerManagement(PM_CONTROL_REG) & ~PM_SOUND_MUTE ) | PM_SOUND_AMP );
  SOUND_CR = SOUND_ENABLE | SOUND_VOL(0x7F);

  irqInit();

  // Start the RTC tracking IRQ
  initClockIRQ();

  SetYtrigger(80);
  irqSet(IRQ_VCOUNT, VcountHandler);
  irqSet(IRQ_VBLANK, VblankHandler);

  irqEnable(IRQ_VBLANK | IRQ_VCOUNT);

  // Keep the ARM7 mostly idle
  COMMAND= cNone;
  while (1) {
    swiWaitForVBlank();

    // temp by maarten
    // Warning from serial.h: These functions (writePowerManagement, readPowerManagement)
    // use the SPI chain, and are thus 'critical'
    // sections, make sure to disable interrupts during the call if you've
    // got a VBlank IRQ polling the touch screen, etc...
    switch( COMMAND ) {
      case cNone:
        // do nothing
        break;
      case cPowerOff:
        REG_IME= IME_DISABLE; // Disable all interrupts
        writePowerManagement(PM_CONTROL_REG, readPowerManagement(PM_CONTROL_REG) | PM_POWER_DOWN );
        REG_IME= IME_ENABLE;  // Enable all interrupts not masked out in REG_IE
        break;
      case cLightBotOn:
        REG_IME= IME_DISABLE; // Disable all interrupts
        writePowerManagement(PM_CONTROL_REG, readPowerManagement(PM_CONTROL_REG) | PM_BACKLIGHT_BOTTOM );
        REG_IME= IME_ENABLE;  // Enable all interrupts not masked out in REG_IE
        break;
      case cLightBotOff:
        REG_IME= IME_DISABLE; // Disable all interrupts
        writePowerManagement(PM_CONTROL_REG, readPowerManagement(PM_CONTROL_REG) & ~PM_BACKLIGHT_BOTTOM );
        REG_IME= IME_ENABLE;  // Enable all interrupts not masked out in REG_IE
        break;
      case cLightTopOn:
        REG_IME= IME_DISABLE; // Disable all interrupts
        writePowerManagement(PM_CONTROL_REG, readPowerManagement(PM_CONTROL_REG) | PM_BACKLIGHT_TOP );
        REG_IME= IME_ENABLE;  // Enable all interrupts not masked out in REG_IE
        break;
      case cLightTopOff:
        REG_IME= IME_DISABLE; // Disable all interrupts
        writePowerManagement(PM_CONTROL_REG, readPowerManagement(PM_CONTROL_REG) & ~PM_BACKLIGHT_TOP );
        REG_IME= IME_ENABLE;  // Enable all interrupts not masked out in REG_IE
        break;
      default:
        // should not occur
        break;
    }
    COMMAND= cNone;
  }
}


