/*
   Watch6 ARM9 Code: main
   Maarten Pennings - 2008 jan 5
   Based on Watch5, but addapted to nds-examples-20071023
   Split-off of queue into separate file. Added key handling.
*/

#include <stdio.h>
#include "nds.h"
#include "fragments.h"
#include "queue.h"

char* Months[] = { "dummy", "jan", "feb", "mar", "apr", "may", "jun", "jul", "aug", "sep", "oct", "nov", "dec" };
char* Weekdays[] = { "sun", "mon", "tue", "wed", "thu", "fri", "sat" };

enum Commands { cNone, cPowerOff, cLightBotOn, cLightBotOff, cLightTopOn, cLightTopOff };
#define COMMAND (IPC->mailData)

enum Lights { lBoth, lTop, lNone, lBot };
enum Chime  { chNone, chHour, chQuarter, chMinute, chQuartMin };


// Definitions for new sound transfer (playSoundsEx)
// =================================================
#include <nds/arm9/cache.h>
#include <string.h>

// The Arm9 sends sound commands to the Arm7.
// It takes some time for the Arm7 to catch up with the Arm9 (asynchronous).
// Each processor records a bit (per channel) indicating the execution of the command.
// The Arm9 records them in ARM9_ALTBITS, and the Arm7 in ARM7_ALTBITS.
// Each processor is the only writer to that variable, but the ARM7_ALTBITS is also read by the Arm9.
// Initially, the bit (for a channel) on the Arm9 is set the same as on the Arm7.
// When the Arm9 sends a command, it flips its local bit.
// When the Arm7 has received and executed the command it also flips its local bit.
// So, when the bits are equal, the Arm7 has caught up with the Arm9.
// Note however, that "executing" means instructing the HW to play the sound fragment.
// Since playing a sound fragment may take an arbitrary long time
// (depending on the size of the PCM block),
// the Arm7 also tells whether it is still playing the fragment (ARM7_BUSY).
uint16 AltBits;
#define ARM9_ALTBITS         (AltBits              )
#define ARM9_ALTBIT(n)       (AltBits      & BIT(n))
#define ARM7_ALTBITS         (IPC->battery         )
#define ARM7_ALTBIT(n)       (IPC->battery & BIT(n))

#define CHANNEL_REQUEST(n)   (ARM9_ALTBIT(n)!=ARM7_ALTBIT(n))

#define ARM7_BUSYBITS        (IPC->aux             )
#define ARM7_BUSYBIT(n)      (IPC->aux     & BIT(n))

#define CHANNEL_FREE(n)      (!CHANNEL_REQUEST(n) && !ARM7_BUSYBIT(n))

TransferSound SndEx;

void playSoundsEx( pTransferSoundData s0, pTransferSoundData s1,  pTransferSoundData s2 ) {
  // Always sends three sounds, but when a sound is NULL, its len is set to 0
  SndEx.count = 3;
  if( s0==NULL ) SndEx.data[0].len=0; else memcpy( &SndEx.data[0], s0, sizeof(TransferSoundData) );
  if( s1==NULL ) SndEx.data[1].len=0; else memcpy( &SndEx.data[1], s1, sizeof(TransferSoundData) );
  if( s2==NULL ) SndEx.data[2].len=0; else memcpy( &SndEx.data[2], s2, sizeof(TransferSoundData) );
  DC_FlushRange( &SndEx, sizeof(TransferSound) );
  // Record in ARM9_ALTBITS which commands we have send out to the Arm7
  if( s0!=NULL ) ARM9_ALTBITS ^= BIT(0);
  if( s1!=NULL ) ARM9_ALTBITS ^= BIT(1);
  if( s2!=NULL ) ARM9_ALTBITS ^= BIT(2);
  IPC->soundData = &SndEx;
}


// Filling queue
// =============
void qAddTime( int hours, int minutes ) {
#define A(id) qAdd( &aFragments[FRAGMENT_##id] )
#define N(i)  qAdd( &aFragments[i] )
  hours= hours % 12;
  if( minutes==0 ) {
    // HetIs <hours> Uur
    A(hetis); N(hours); A(uur);
  } else if( minutes<15 ) {
    // HetIs <minutes> Over <hours>
    A(hetis); N(minutes); A(over); N(hours);
  } else if( minutes==15 ) {
    // HetIs Kwart Over <hours>
    A(hetis); A(kwart); A(over); N(hours);
  } else if( minutes<30 ) {
    // HetIs <30-minutes> Voor Half <hours+1>
    A(hetis); N(30-minutes); A(voor); A(half); N(hours+1);
  } else if( minutes==30 ) {
    // HetIs Half <hours+1>
    A(hetis); A(half); N(hours+1);
  } else if( minutes<45 ) {
    // HetIs <minutes-30> Over Half <hours+1>
    A(hetis); N(minutes-30); A(over); A(half); N(hours+1);
  } else if( minutes==45 ) {
    // HetIs Kwart Voor <hours+1>
    A(hetis); A(kwart); A(voor); N(hours+1);
  } else {
    // HetIs <60-minutes> Voor <hours+1>
    A(hetis); N(60-minutes); A(voor); N(hours+1);
  }
}


// Definitions for GL helpers
// ==========================
void DrawKite(float z, float width, float height)
{
  glBegin(GL_QUADS);
  glVertex3f( -width, 0       , z);
  glVertex3f( 0     , -width  , z);
  glVertex3f( +width, 0       , z);
  glVertex3f( 0     , +height , z);
  glEnd();
}


void DrawQuad(float z, float width, float height)
{
  glBegin(GL_QUADS);
  glVertex3f( -width, -height , z);
  glVertex3f( +width, -height , z);
  glVertex3f( +width, +height , z);
  glVertex3f( -width, +height , z);
  glEnd();
}


void InitGraphics() {
  // Turn on everything
  powerON(POWER_ALL);

  //put 3D on top
  lcdMainOnTop();

  // Setup the Main screen for 3D
  videoSetMode(MODE_0_3D);

  // IRQ basic setup
  irqInit();
  irqSet(IRQ_VBLANK, 0);

  // Set our viewport to be the same size as the screen
  glViewport(0,0,255,191);

  // Specify the Clear Color and Depth
  glClearColor(0,0,0,31);
  glClearDepth(0x7FFF);

  consoleDemoInit();

  // Reset the screen and setup the view
  glInit();

  //ds specific, several attributes can be set here
  glPolyFmt(POLY_ALPHA(31) | POLY_CULL_NONE);

  // Set the current matrix to be the model matrix
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(70, 256.0 / 192.0, 0.1, 100);
  gluLookAt(  0.0, 0.0, 3.5,    //camera possition
              0.0, 0.0, 0.0,    //look at
              0.0, 1.0, 0.0);   //up
}


void DrawGlScene( int hours, int minutes, int seconds ) {
  int h;

  // Draw second hand
  glPushMatrix();
  glColor3f(0,0,1);
  glRotateZ(-seconds * 360 / 60);
  DrawKite(0.03,0.09,2.1);
  glPopMatrix(1);

  // Draw minute hand
  glPushMatrix();
  glColor3f(0,1,0);
  glRotateZ(- (minutes*60 + seconds) / 10);
  DrawKite(0.02,0.15,1.85);
  glPopMatrix(1);

  // Draw hourhand
  glPushMatrix();
  glColor3f(1,0,0);
  glRotateZ(- (hours*3600 + minutes*60 + seconds) / 120);
  DrawKite(0.01,0.20,1.50);
  glPopMatrix(1);

  for( h=0; h<12; h++ ) {
    // Draw hour tick
    glPushMatrix();
    glColor3f(1,1,1);
    glRotateZ(-h * 360 / 12);
    glTranslatef(0,1.9,0);
    DrawQuad(0.00,(h%3==0?0.10:0.03),0.15);
    glPopMatrix(1);
  }
}


// Main Arm9
// =========
int main(void) {
  int bTikTak= 1;
  enum Chime eChime= chQuartMin;
  enum Lights eLights= lBoth;
  int bPower= 1;
  int prevsec;
  int prevmin;
  int prevtouched;

  qInit();         // no pending fragments to play
  InitFragments(); // set up the array of sound fragments (see fragments.c)
  InitGraphics();  // prepare for GL

  // Wait till we get the first read-out
  prevsec= IPC->time.rtc.seconds;
  while( prevsec==IPC->time.rtc.seconds ) {
    prevsec=IPC->time.rtc.seconds;
  }
  prevsec= IPC->time.rtc.seconds;
  prevmin= IPC->time.rtc.minutes;
  prevtouched= 0;

  // Set current status of alternating bits
  ARM9_ALTBITS= 0; // Arbitrary 'alternating bit' settings for all channels (all 0) -- must be the same on Arm7

  printf( "\x1b[1;3H" );
  printf( "Watch6 - Maarten Pennings" );

  while(1) {
    pTransferSoundData snd0, snd1, snd2;

    // Time data handling
    int year=    IPC->time.rtc.year+2000;
    int month=   IPC->time.rtc.month;
    int day=     IPC->time.rtc.day;
    int weekday= IPC->time.rtc.weekday;
    int hours=   IPC->time.rtc.hours;
    int minutes= IPC->time.rtc.minutes;
    int seconds= IPC->time.rtc.seconds;

    // Key handling
    scanKeys();
    if( keysDown() & KEY_X )
    {
      bTikTak= !bTikTak;
    }
    if( keysDown() & KEY_Y )
    {
      if( eChime==chQuartMin ) eChime=chNone; else eChime++;
    }
    if( keysDown() & KEY_A )
    {
      bPower= !bPower;
      if( !bPower ) COMMAND= cPowerOff;
    }
    if( keysDown() & KEY_B )
    {
      if( eLights==lBoth ) { eLights=lTop; COMMAND=cLightBotOff; }
      else if( eLights==lTop ) { eLights=lNone; COMMAND=cLightTopOff; }
      else if( eLights==lNone ) { eLights=lBot; COMMAND=cLightBotOn; }
      else { eLights=lBoth; COMMAND=cLightTopOn; }
    }
    if( keysDown() & KEY_TOUCH )
    {
      // Choice: only one time telling request is queued at a time
      if( qEmpty() ) qAddTime(hours, minutes);
    }

    // Status printing
    printf( "\x1b[2;3H" );
    printf( "%i-%s-%i (%s) ", year, Months[month], day, Weekdays[weekday] );
    printf( "%2i:%02i:%02i \n", hours, minutes, seconds );
    printf( "\x1b[3;1H" );
    printf( "Touch screen for time telling" );
    printf( "\x1b[10;1H" );
    printf( "X:tik/tak:%s", bTikTak ? "on  " : "off " );
    printf( "\x1b[10;15H" );
    printf( "Y:chime:%s", eChime==chNone ? "none" : eChime==chHour ? "hour" : eChime==chQuarter ? "qhr " : eChime==chMinute ? "min " : "qmin" );
    printf( "\x1b[11;1H" );
    printf( "A:power:%s", bPower ? "on  " : "off " );
    printf( "\x1b[11;15H" );
    printf( "B:light:%s", eLights==lBoth ? "both" : eLights==lTop ? "top " : eLights==lBot ? "bot " : "none" );

    // Sound handling
    snd0= snd1= snd2= NULL;
    // Recall that the while loop is active 60x per second. Are we at a full second? If yes play either tik or tak.
    if( bTikTak )
      if( prevsec!=seconds ) snd0= seconds%2==0 ? &aFragments[FRAGMENT_tik] : &aFragments[FRAGMENT_tak] ;
    // Play a chime at "important" moments.
    switch( eChime ) {
      case chNone:     break;
      case chHour:     if( prevmin!=minutes && minutes==0    ) snd1=&aFragments[FRAGMENT_chime]; break;
      case chQuarter:  if( prevmin!=minutes && minutes%15==0 ) snd1=&aFragments[FRAGMENT_chime]; break;
      case chMinute:   if( prevsec!=seconds && seconds==0    ) snd1=&aFragments[FRAGMENT_chime]; break;
      case chQuartMin: if( prevsec!=seconds && seconds%15==0 ) snd1=&aFragments[FRAGMENT_chime]; break;
    }
    // Do we have a pending request for time telling (queue not empty)?
    if( !qEmpty() && CHANNEL_FREE(2) ) snd2=qGet();
    playSoundsEx(snd0,snd1,snd2);

    // GL Rendering
    glPushMatrix();
    DrawGlScene(hours,minutes,seconds);
    glPopMatrix(1);
    glFlush(0);

    prevmin=minutes;
    prevsec=seconds;
  }
}

